import zipfile
import zipimport
from pathlib import Path
from sys import modules, stderr


def activate(path=None):
    path = Path(path) if path else Path(__file__).parent
    path_str = path.absolute().as_posix()
    if zipfile.is_zipfile(path_str):
        try:
            _tmp = zipimport.zipimporter(path_str).load_module("ensure_zipapps")
            modules.pop(_tmp.__name__, None)
            del _tmp
            return True
        except ImportError as err:
            stderr.write(f"WARNING: activate failed for {err!r}\n")
            raise err



import asyncio
import pickle
import socket
import sys
from abc import ABC, abstractmethod
from concurrent.futures import ThreadPoolExecutor
from logging import LogRecord
from logging.handlers import SocketHandler
from typing import Any, Callable, Dict, Literal, Optional, Union

__all__ = [
    "IPCEncoder",
    "JSONEncoder",
    "PickleEncoder",
    "SocketLogHandlerEncoder",
    "SocketServer",
    "SocketClient",
    "find_free_port",
]


class IPCEncoder(ABC):
    """An abstract base class for all encoders; implementing the necessary communication protocol requires only the definition of two abstract methods. Be mindful that varying header lengths will impact the maximum packaging size."""

    # HEAD_LENGTH: Package length
    # 1: 256 B, 2: 64 KB, 3: 16 MB, 4: 4 GB, 5: 1 TB, 6: 256 TB
    HEAD_SIZE = 4
    BYTEORDER: Literal["little", "big"] = "big"

    @abstractmethod
    def _dumps(self, raw: Any) -> bytes:
        raise NotImplementedError

    @abstractmethod
    def _loads(self, raw: bytes) -> Any:
        raise NotImplementedError

    def dumps(self, raw: Any) -> bytes:
        result = self._dumps(raw)
        head = len(result).to_bytes(self.HEAD_SIZE, self.BYTEORDER)
        return head + result

    def loads(self, raw: Any) -> bytes:
        return self._loads(raw)

    def get_size(self, head: bytes):
        return int.from_bytes(head, self.BYTEORDER)


class JSONEncoder(IPCEncoder):
    _DUMP_KWARGS: Dict = {}
    _LOAD_KWARGS: Dict = {}
    _ENCODING = "utf-8"

    def __init__(self):
        import json

        self._encoder = json.dumps
        self._decoder = json.loads
        super().__init__()

    def _dumps(self, raw: Any) -> bytes:
        return self._encoder(raw, **self._DUMP_KWARGS).encode(self._ENCODING)

    def _loads(self, raw: bytes) -> Any:
        return self._decoder(raw.decode(self._ENCODING), **self._LOAD_KWARGS)


class PickleEncoder(IPCEncoder):
    _DUMP_KWARGS: Dict = {}
    _LOAD_KWARGS: Dict = {}

    def __init__(self):
        super().__init__()

    def _dumps(self, raw: Any) -> bytes:
        return pickle.dumps(raw, **self._DUMP_KWARGS)

    def _loads(self, raw: bytes) -> Any:
        return pickle.loads(raw, **self._LOAD_KWARGS)


class SocketLogHandlerEncoder(IPCEncoder):
    """For a practical demonstration, refer to the test code: morebuiltins/ipc.py:_test_ipc_logging.

    Demo::

        async def _test_ipc_logging():
            import logging

            host = "127.0.0.1"
            port = 8090
            async with SocketServer(host=host, port=port, encoder=SocketLogHandlerEncoder()):
                # socket logger demo
                # ==================
                logger = logging.getLogger("test_logger")
                logger.setLevel(logging.DEBUG)
                h = SocketHandler(host, port)
                h.setLevel(logging.DEBUG)
                logger.addHandler(h)
                logger.info("test socket")
                # ==================

                # ensure test case
                await asyncio.sleep(0.1)
                assert pickle.loads(h.sock.recv(100000)[4:])["name"] == logger.name
    And provide a simple implementation for generating logs for coroutine code with Client usage.
    """

    _DUMP_KWARGS: Dict = {"protocol": 1}
    _LOAD_KWARGS: Dict = {}

    def __init__(self):
        super().__init__()

    def makePickle(self, d: Union[dict, LogRecord]):
        if isinstance(d, LogRecord):
            msg = d.getMessage()
            d = dict(d.__dict__)
            d["msg"] = msg
            d["args"] = None
            d["exc_info"] = None
            d.pop("message", None)
        return pickle.dumps(d, **self._DUMP_KWARGS)

    def _dumps(self, raw: Union[LogRecord, dict]) -> bytes:
        return self.makePickle(raw)

    def _loads(self, raw: bytes) -> Any:
        return pickle.loads(raw, **self._LOAD_KWARGS)


class SocketServer:
    """To see an example in action, view the test code: morebuiltins/ipc.py:_test_ipc.

    Demo::

        async def test_client(host="127.0.0.1", port=8090, encoder=None, cases=None):
            async with SocketClient(host=host, port=port, encoder=encoder) as c:
                for case in cases:
                    await c.send(case)
                    response = await c.recv()
                    if globals().get("print_log"):
                        print("[Client]", "send:", repr(case), "=>", "recv:", repr(response))
                    assert case == response or str(case) == response, [case, response]
                await c.send("[shutdown server]")


        async def _test_ipc():
            import platform

            JSONEncoder._DUMP_KWARGS["default"] = str
            for enc, cases in [
                [PickleEncoder, [123, "123", None, {"a"}, ["a"], ("a",), {"a": 1}]],
                [JSONEncoder, [123, "123", None, {"a"}, ["a"], {"a": 1}]],
            ]:
                encoder = enc()
                if platform.system() == "Linux":
                    # test unix domain socket
                    print("Test Linux Unix Domain Socket")
                    host = "/tmp/uds.sock"
                    port = None
                    async with SocketServer(host=host, port=port, encoder=encoder):
                        await test_client(host, port=None, encoder=encoder, cases=cases)

                # test socket
                host = "127.0.0.1"
                port = 8090
                async with SocketServer(host=host, port=port, encoder=encoder):
                    await test_client(host="127.0.0.1", port=8090, encoder=encoder, cases=cases)
    """

    def __init__(
        self,
        host: str = "127.0.0.1",
        port: int = 8090,
        handler: Optional[Callable] = None,
        encoder: Optional[IPCEncoder] = None,
        connect_kwargs: Optional[dict] = None,
        start_callback: Optional[Callable] = None,
        end_callback: Optional[Callable] = None,
    ):
        if port is None and sys.platform == "win32":
            raise SystemError("not support UDS(unix domain socket) on win32 platform")
        self.host = host
        self.port = port
        self.handler = handler or self.default_handler
        self.encoder: IPCEncoder = encoder or PickleEncoder()
        self.connect_kwargs = connect_kwargs or {}
        self.start_callback = start_callback
        self.end_callback = end_callback

        self.server: Optional[asyncio.base_events.Server] = None
        self._shutdown_ev: Optional[asyncio.Event] = None
        self._default_executor = ThreadPoolExecutor()

    async def default_handler(self, item: Any):
        if globals().get("print_log"):
            print("[Server] recv:", repr(item), "=>", "send:", repr(item), flush=True)
        if item == "[shutdown server]":
            await self.close()
        else:
            return self.encoder.dumps(item)

    async def connect_callback(
        self,
        reader: asyncio.StreamReader,
        writer: asyncio.StreamWriter,
    ) -> None:
        try:
            head_size = self.encoder.HEAD_SIZE
            need_await = asyncio.iscoroutinefunction(self.handler)
            while (
                not (self._shutdown_ev and self._shutdown_ev.is_set())
                and self.is_serving()
                and not reader.at_eof()
            ):
                head = await reader.read(head_size)
                if len(head) < head_size:
                    break
                content_length = self.encoder.get_size(head)
                # read the whole package
                head = await reader.read(content_length)
                while len(head) < content_length:
                    head = head + await reader.read(content_length - len(head))
                item = self.encoder._loads(head)
                if need_await:
                    result = await self.handler(item)
                else:
                    result = asyncio.get_running_loop().run_in_executor(
                        self._default_executor, self.handler, item
                    )
                if isinstance(result, bytes):
                    writer.write(result)
                    await writer.drain()
        finally:
            writer.close()
            await writer.wait_closed()

    async def close(self):
        if self.is_serving():
            self.server.close()
            await self.server.wait_closed()
            self._shutdown_ev.set()
            if self.end_callback:
                if asyncio.iscoroutinefunction(self.end_callback):
                    await self.end_callback()
                else:
                    await asyncio.get_running_loop().run_in_executor(
                        self._default_executor, self.end_callback
                    )

    async def start(self):
        self._shutdown_ev = asyncio.Event()
        if self.port:
            self.server = await asyncio.start_server(
                self.connect_callback,
                host=self.host,
                port=self.port,
                **self.connect_kwargs,
            )
            await self.server.start_serving()
        else:
            self.server = await asyncio.start_unix_server(
                self.connect_callback, path=self.host, **self.connect_kwargs
            )
        if self.start_callback:
            if asyncio.iscoroutinefunction(self.start_callback):
                await self.start_callback()
            else:
                await asyncio.get_running_loop().run_in_executor(
                    self._default_executor, self.start_callback
                )

    def is_serving(self):
        return self.server and self.server.is_serving()

    async def wait_closed(self):
        if self._shutdown_ev:
            await self._shutdown_ev.wait()

    def shutdown(self):
        "sync close"
        if self.is_serving():
            self.server.get_loop().call_soon_threadsafe(self._shutdown_ev.set)

    async def __aenter__(self):
        await self.start()
        return self

    async def __aexit__(self, *_errors):
        await self.close()


class SocketClient:
    def __init__(
        self,
        host: str = "127.0.0.1",
        port: int = 8090,
        encoder: Optional[IPCEncoder] = None,
        connect_kwargs: Optional[dict] = None,
    ):
        self.host = host
        self.port = port
        self.encoder: IPCEncoder = encoder or PickleEncoder()
        self.connect_kwargs = connect_kwargs or {}

    async def __aenter__(self):
        if self.port:
            self.reader, self.writer = await asyncio.open_connection(
                self.host, self.port, **self.connect_kwargs
            )
        else:
            self.reader, self.writer = await asyncio.open_unix_connection(
                path=self.host, **self.connect_kwargs
            )
        self.lock = asyncio.Lock()
        return self

    async def __aexit__(self, *_):
        if self.writer:
            self.writer.close()
            await self.writer.wait_closed()

    async def send(self, item: Any):
        assert self.writer, "use `async with`"
        async with self.lock:
            self.writer.write(self.encoder.dumps(item))
            await self.writer.drain()

    async def recv(self) -> Any:
        reader = self.reader
        assert reader, "use `async with`"
        async with self.lock:
            assert not reader.at_eof()
            head = await reader.read(self.encoder.HEAD_SIZE)
            if len(head) < self.encoder.HEAD_SIZE:
                raise RuntimeError()
            content_length = self.encoder.get_size(head)
            head = await reader.read(content_length)
            while len(head) < content_length:
                head = head + await reader.read(content_length - len(head))
            return self.encoder.loads(head)


def find_free_port(host="127.0.0.1", port=0):
    """Finds and returns an available port number.

    Parameters:
    - host: The host address to bind, default is "127.0.0.1".
    - port: The port number to attempt binding, default is 0 (for OS allocation).

    Returns:
    - If a free port is found, it returns the port number; otherwise, returns None.

    Demo:

    >>> free_port = find_free_port()
    >>> isinstance(free_port, int)
    True
    """
    try:
        with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
            s.bind((host, port))
            return s.getsockname()[1]
    except socket.error:
        pass


async def test_client(host="127.0.0.1", port=8090, encoder=None, cases=None):
    async with SocketClient(host=host, port=port, encoder=encoder) as c:
        for case in cases:
            await c.send(case)
            response = await c.recv()
            if globals().get("print_log"):
                print("[Client]", "send:", repr(case), "=>", "recv:", repr(response))
            assert case == response or str(case) == response, [case, response]
        await c.send("[shutdown server]")


async def _test_ipc():
    import platform

    JSONEncoder._DUMP_KWARGS["default"] = str
    for enc, cases in [
        [PickleEncoder, [123, "123", None, {"a"}, ["a"], ("a",), {"a": 1}]],
        [JSONEncoder, [123, "123", None, {"a"}, ["a"], {"a": 1}]],
    ]:
        encoder = enc()
        if platform.system() == "Linux":
            # test unix domain socket
            print("Test Linux Unix Domain Socket")
            host = "/tmp/uds.sock"
            port = None
            async with SocketServer(host=host, port=port, encoder=encoder):
                await test_client(host, port=None, encoder=encoder, cases=cases)

        # test socket
        host = "127.0.0.1"
        port = 8090
        async with SocketServer(host=host, port=port, encoder=encoder):
            await test_client(host="127.0.0.1", port=8090, encoder=encoder, cases=cases)


async def _test_ipc_logging():
    import logging

    host = "127.0.0.1"
    port = 8090
    assert find_free_port(host, port) == port
    async with SocketServer(host=host, port=port, encoder=SocketLogHandlerEncoder()):
        assert find_free_port(host, port) is None
        # socket logger demo
        # ==================
        logger = logging.getLogger("test_logger")
        logger.setLevel(logging.DEBUG)
        h = SocketHandler(host, port)
        h.setLevel(logging.DEBUG)
        logger.addHandler(h)
        logger.info("test socket")
        # ==================

        # ensure test case
        await asyncio.sleep(0.1)
        assert pickle.loads(h.sock.recv(100000)[4:])["name"] == logger.name
        h.sock.close()


def test():
    globals().setdefault("print_log", True)  # local test show logs
    for function in [_test_ipc, _test_ipc_logging]:
        asyncio.get_event_loop().run_until_complete(function())


if __name__ == "__main__":
    test()



import asyncio
import re
import sys
import time
import typing
from collections import namedtuple
from pathlib import Path

__all__ = ["ProxyChecker"]


proxy_info = namedtuple("proxy_info", "proxy ok cost error")


class ProxyChecker(object):
    """A command line toolkit to check available proxies.

    1. clipboard usage:
        > input-text from clipboard, and set the result to clipboard. `-l 2` means try 2 loops
            > python -m morebuiltins.cmd.proxy_checker -c -l 2

    2. input-file output-file usage:
        > input-text from file, and set the result to output-file
            > python -m morebuiltins.cmd.proxy_checker -i input.txt -o output.txt
        > output to stdout
            > python -m morebuiltins.cmd.proxy_checker -i input.txt

    3. stdin usage:
        > cat file.txt | python -m morebuiltins.cmd.proxy_checker > stdout.txt
        > cat file.txt | python -m morebuiltins.cmd.proxy_checker -o output.txt

    python -m morebuiltins.cmd.proxy_checker -h
        options:
        -h, --help            show this help message and exit
        -t TIMEOUT, --timeout TIMEOUT
                                timeout of each request
        -l LOOP, --loop LOOP  Loop the test to validate the successful results each time until the desired number of iterations is reached.
        --retry RETRY         retry times
        -n CONCURRENCY, --concurrency CONCURRENCY
                                concurrency
        -i INPUT_FILE, --input-file INPUT_FILE
                                input text file path
        -o OUTPUT_FILE, --output-file OUTPUT_FILE
                                output text file path
        -c, --from-clipboard  text from clipboard, ignore -i. if --output-file not set, output to clipboard
        -q, --quiet           mute the progress in stderr


    """

    target = re.compile(rb"HTTP/1.1 20\d [^\r\n]+\r\n")
    request = b"GET http://www.gstatic.com/generate_204 HTTP/1.1\r\nHost: www.gstatic.com\r\nConnection: close\r\n\r\n"
    ip_regex = re.compile(r"(\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3})(?:\s+|:)(\d+)")

    def __init__(self, concurrency=20, timeout=3.0, retry=0):
        self.sem = asyncio.Semaphore(concurrency)
        self.timeout = timeout
        self.retry = retry

    @classmethod
    def find_proxies(cls, text: str):
        return [f"http://{host}:{port}" for host, port in cls.ip_regex.findall(text)]

    async def check_proxies(
        self, proxy_list: typing.List[str]
    ) -> typing.AsyncGenerator[proxy_info, None]:
        tasks = [asyncio.create_task(self.check(proxy)) for proxy in proxy_list]
        for task in asyncio.as_completed(tasks):
            item = await task
            yield item

    async def send_request(self, host: str, port: int):
        try:
            writer = None
            reader, writer = await asyncio.open_connection(host, port)
            writer.write(self.request)
            await writer.drain()
            return await reader.readline()
        finally:
            if writer:
                writer.close()
                await writer.wait_closed()

    async def check(self, proxy: str):
        ok = cost = error = None
        for _ in range(self.retry + 1):
            try:
                start = time.time()
                match = self.ip_regex.search(proxy)
                if not match:
                    error = 'ValueError("invalid proxy")'
                    break
                host, port = match.groups()
                async with self.sem:
                    start = time.time()
                    result = await asyncio.wait_for(
                        self.send_request(host, int(port)), timeout=self.timeout
                    )
                ok = bool(self.target.search(result))
                if ok:
                    break
            except Exception as e:
                error = repr(e)
            finally:
                cost = round(time.time() - start, 3)

        return proxy_info(proxy, ok, cost, error)


# 使用示例
async def main():
    import argparse

    parser = argparse.ArgumentParser()
    parser.add_argument(
        "-t", "--timeout", type=float, default=3.0, help="timeout of each request"
    )
    parser.add_argument(
        "-l",
        "--loop",
        type=int,
        default=1,
        help="Loop the test to validate the successful results each time until the desired number of iterations is reached.",
    )
    parser.add_argument("--retry", type=int, default=0, help="retry times")
    parser.add_argument("-n", "--concurrency", type=int, default=20, help="concurrency")
    parser.add_argument(
        "-r",
        "--max-result",
        type=int,
        default=0,
        help="quick return when result count >= max-result",
    )
    parser.add_argument("-i", "--input-file", default="", help="input text file path")
    parser.add_argument("-o", "--output-file", default="", help="output text file path")
    parser.add_argument(
        "-c",
        "--from-clipboard",
        action="store_true",
        help="text from clipboard, ignore -i. if --output-file not set, output to clipboard",
    )
    parser.add_argument(
        "-q", "--quiet", action="store_true", help="mute the progress in stderr"
    )
    parser.add_argument(
        "-u",
        "--url",
        default="http://www.gstatic.com/generate_204",
        help="default check url",
    )
    parser.add_argument(
        "-ii", "--stdin", action="store_true", help="set stdin as input"
    )
    args = parser.parse_args()
    quiet = args.quiet
    text = ""
    output_file = None
    from_clipboard = args.from_clipboard
    target_path = Path(args.output_file) if args.output_file else None
    if args.input_file:
        text = Path(args.input_file).read_bytes().decode("utf-8", "ignore")
    elif args.stdin:
        null_lines = 0
        for i in sys.stdin:
            if not i.strip():
                null_lines += 1
            if null_lines > 3:
                break
            text += i
        if not args.output_file:
            output_file = sys.stdout
            quiet = True
    else:
        from morebuiltins.utils import Clipboard

        text = Clipboard.paste()
        target_path = target_path or Clipboard.copy
    pc = ProxyChecker(
        concurrency=args.concurrency, timeout=args.timeout, retry=args.retry
    )
    proxies = pc.find_proxies(text)
    result_list = []
    try:
        if isinstance(target_path, Path):
            target_path = Path(args.output_file)
            target_path.parent.mkdir(parents=True, exist_ok=True)
            output_file = open(target_path, "w")
        for loop in range(args.loop):
            proxies = set(result_list or proxies)
            result_list.clear()
            done = 0
            oks = 0
            last_loop = args.loop == loop + 1
            async for item in pc.check_proxies(proxies):
                done += 1
                if item.ok:
                    oks += 1
                    result_list.append(item.proxy)
                    if (
                        last_loop
                        and args.max_result
                        and len(result_list) >= args.max_result
                    ):
                        break
                if not quiet:
                    print(
                        f"[{loop+1}] {done} / {len(proxies)} (oks={oks})".ljust(21),
                        item[0].ljust(28),
                        *item[1:],
                        flush=True,
                        file=sys.stderr,
                    )
        if result_list or from_clipboard:
            result = "\n".join(result_list)
            if output_file:
                print(result, file=output_file, flush=True)
            elif target_path:
                target_path(result)
    finally:
        if output_file and output_file is not sys.stdout:
            output_file.close()


if __name__ == "__main__":
    asyncio.run(main())



import asyncio
import importlib.util
import inspect
import json
import os
import re
import time
from collections import deque
from concurrent.futures import ThreadPoolExecutor
from functools import wraps
from itertools import chain
from logging.handlers import TimedRotatingFileHandler
from pathlib import Path
from threading import Lock, Semaphore
from typing import Callable, Coroutine, Dict, Optional, OrderedDict, Set, Union
from weakref import WeakSet

__all__ = [
    "lru_cache_ttl",
    "threads",
    "bg_task",
    "NamedLock",
    "FuncSchema",
    "InlinePB",
    "SizedTimedRotatingFileHandler",
    "get_type_default",
    "func_cmd",
    "file_import",
    "RotatingFileWriter",
]


def lru_cache_ttl(
    maxsize: int,
    ttl: Optional[Union[int, float]] = None,
    controls=False,
    auto_clear=True,
    timer=time.time,
):
    """A Least Recently Used (LRU) cache with a Time To Live (TTL) feature.

    Args:
        maxsize (int): maxsize of cache
        ttl (Optional[Union[int, float]], optional): time to live. Defaults to None.
        controls (bool, optional): set cache/ttl_clean attributes. Defaults to False.
        auto_clear (bool, optional): clear dead cache automatically. Defaults to True.
        timer (callable, optional): Defaults to time.time.

    Returns:
        callable: decorator function

    >>> import time
    >>> # test ttl
    >>> values = [1, 2]
    >>> @lru_cache_ttl(1, 0.1)
    ... def func1(i):
    ...     return values.pop(0)
    >>> [func1(1), func1(1), time.sleep(0.11), func1(1)]
    [1, 1, None, 2]
    >>> # test maxsize
    >>> values = [1, 2, 3]
    >>> func = lambda i: values.pop(0)
    >>> func1 = lru_cache_ttl(2)(func)
    >>> [func1(i) for i in [1, 1, 1, 2, 2, 2, 3, 3, 3]]
    [1, 1, 1, 2, 2, 2, 3, 3, 3]
    >>> # test auto_clear=True, with controls
    >>> values = [1, 2, 3, 4]
    >>> func = lambda i: values.pop(0)
    >>> func1 = lru_cache_ttl(5, 0.1, controls=True, auto_clear=True)(func)
    >>> [func1(1), func1(2), func1(3)]
    [1, 2, 3]
    >>> time.sleep(0.11)
    >>> func1(3)
    4
    >>> len(func1.cache)
    1
    >>> # test auto_clear=False
    >>> values = [1, 2, 3, 4]
    >>> @lru_cache_ttl(5, 0.1, controls=True, auto_clear=False)
    ... def func1(i):
    ...     return values.pop(0)
    >>> [func1(1), func1(2), func1(3)]
    [1, 2, 3]
    >>> time.sleep(0.11)
    >>> func1(3)
    4
    >>> len(func1.cache)
    3
    """
    cache: OrderedDict = OrderedDict()
    move_to_end = cache.move_to_end
    popitem = cache.popitem
    next_clear_ts = timer()

    def ttl_clean(expire):
        for k, v in tuple(cache.items()):
            if v[1] <= expire:
                cache.pop(k, None)

    def decorator(func):
        @wraps(func)
        def wrapped(*args, **kwargs):
            nonlocal next_clear_ts
            now = timer()
            if ttl is None:
                expire_cleared = True
            else:
                expire_cleared = auto_clear and now > next_clear_ts
                if expire_cleared:
                    ttl_clean(now)
                    next_clear_ts = now + ttl
            # key = _make_key(args, kwargs, False)
            key = hash(tuple(chain(args, kwargs.items())))
            if key in cache:
                if expire_cleared is True or now < cache[key][1]:
                    # 1. key in cache; 2. cache is valid
                    # set newest, return result
                    move_to_end(key)
                    return cache[key][0]
            # call function
            result = func(*args, **kwargs)
            # ensure size
            while len(cache) >= maxsize:
                popitem(last=False)
            if ttl is None:
                # no need save expired ts
                cache[key] = (result,)
            else:
                # setitem(key, (result, now + ttl))
                cache[key] = (result, timer() + ttl)
            # {key: (result, expired_ts)}
            return result

        if controls:
            setattr(wrapped, "cache", cache)
            setattr(wrapped, "ttl_clean", ttl_clean)
        return wrapped

    return decorator


def threads(n: Optional[int] = None, executor_class=None, **kws):
    """Quickly convert synchronous functions to be concurrency-able. (similar to madisonmay/Tomorrow)

    >>> @threads(10)
    ... def test(i):
    ...     time.sleep(i)
    ...     return i
    >>> start = time.time()
    >>> tasks = [test(i) for i in [0.1] * 5]
    >>> len(test.pool._threads)
    5
    >>> len(test.tasks)
    5
    >>> for i in tasks:
    ...     i.result() if hasattr(i, 'result') else i
    0.1
    0.1
    0.1
    0.1
    0.1
    >>> time.time() - start < 0.2
    True
    >>> len(test.pool._threads)
    5
    >>> len(test.tasks)
    0
    """
    pool = (executor_class or ThreadPoolExecutor)(max_workers=n, **kws)
    tasks: WeakSet = WeakSet()

    def decorator(func):
        @wraps(func)
        def wrapped(*args, **kwargs):
            future = pool.submit(func, *args, **kwargs)
            future.add_done_callback(lambda task: tasks.discard(task))
            tasks.add(future)
            return future

        setattr(wrapped, "pool", pool)
        setattr(wrapped, "tasks", tasks)
        return wrapped

    return decorator


_bg_tasks: Set[asyncio.Task] = set()


def bg_task(coro: Coroutine) -> asyncio.Task:
    """Avoid asyncio free-flying tasks, better to use the new asyncio.TaskGroup to avoid this in 3.11+. https://github.com/python/cpython/issues/91887

    Args:
        coro (Coroutine)

    Returns:
        _type_: Task

    """
    task = asyncio.create_task(coro)
    _bg_tasks.add(task)
    task.add_done_callback(_bg_tasks.discard)
    return task


LockType = Union[Lock, Semaphore, asyncio.Lock, asyncio.Semaphore]


class NamedLock:
    """Reusable named locks, support for timeouts, support for multiple concurrent locks.

    Demo::

        def test_named_lock():
            def test_sync():
                import time
                from concurrent.futures import ThreadPoolExecutor
                from threading import Lock, Semaphore

                def _test1():
                    with NamedLock("_test1", Lock, timeout=0.05) as lock:
                        time.sleep(0.2)
                        return bool(lock)

                with ThreadPoolExecutor(10) as pool:
                    tasks = [pool.submit(_test1) for _ in range(3)]
                    result = [i.result() for i in tasks]
                    assert result == [True, False, False], result
                assert len(NamedLock._SYNC_CACHE) == 1
                NamedLock.clear_unlocked()
                assert len(NamedLock._SYNC_CACHE) == 0

                def _test2():
                    with NamedLock("_test2", lambda: Semaphore(2), timeout=0.05) as lock:
                        time.sleep(0.2)
                        return bool(lock)

                with ThreadPoolExecutor(10) as pool:
                    tasks = [pool.submit(_test2) for _ in range(3)]
                    result = [i.result() for i in tasks]
                    assert result == [True, True, False], result

            def test_async():
                import asyncio

                async def main():
                    async def _test1():
                        async with NamedLock("_test1", asyncio.Lock, timeout=0.05) as lock:
                            await asyncio.sleep(0.2)
                            return bool(lock)

                    tasks = [asyncio.create_task(_test1()) for _ in range(3)]
                    result = [await i for i in tasks]
                    assert result == [True, False, False], result
                    assert len(NamedLock._ASYNC_CACHE) == 1
                    NamedLock.clear_unlocked()
                    assert len(NamedLock._ASYNC_CACHE) == 0

                    async def _test2():
                        async with NamedLock(
                            "_test2", lambda: asyncio.Semaphore(2), timeout=0.05
                        ) as lock:
                            await asyncio.sleep(0.2)
                            return bool(lock)

                    tasks = [asyncio.create_task(_test2()) for _ in range(3)]
                    result = [await i for i in tasks]
                    assert result == [True, True, False], result

                asyncio.get_event_loop().run_until_complete(main())

            test_sync()
            test_async()
    """

    _SYNC_CACHE: Dict[str, LockType] = {}
    _ASYNC_CACHE: Dict[str, LockType] = {}

    def __init__(self, name: str, default_factory: Callable, timeout=None):
        self.name = name
        self.default_factory = default_factory
        self.timeout = timeout
        self.lock = None

    @classmethod
    def clear_unlocked(cls):
        for cache in [cls._SYNC_CACHE, cls._ASYNC_CACHE]:
            for name, lock in list(cache.items()):
                if hasattr(lock, "locked") and not lock.locked():
                    cache.pop(name, None)
                elif isinstance(lock, Semaphore) and (
                    lock._value == 0 or not lock._cond._lock.locked()
                ):
                    cache.pop(name, None)

    def __enter__(self):
        if self.name in self._SYNC_CACHE:
            lock = self._SYNC_CACHE[self.name]
        else:
            lock = self.default_factory()
            self._SYNC_CACHE[self.name] = lock
        if lock.acquire(timeout=self.timeout):
            self.lock = lock
            return self
        else:
            return None

    def __exit__(self, *_):
        if self.lock:
            self.lock.release()

    async def __aenter__(self):
        if self.name in self._ASYNC_CACHE:
            lock = self._ASYNC_CACHE[self.name]
        else:
            lock = self.default_factory()
            self._ASYNC_CACHE[self.name] = lock
        try:
            await asyncio.wait_for(lock.acquire(), timeout=self.timeout)
            self.lock = lock
            return self
        except asyncio.TimeoutError:
            return None

    async def __aexit__(self, *_):
        if self.lock:
            self.lock.release()


class FuncSchema:
    """Parse the parameters and types required by a function into a dictionary, and convert an incoming parameter into the appropriate type.

    >>> def test(a, b: str, /, c=1, *, d=["d"], e=0.1, f={"f"}, g=(1, 2), h=True, i={1}, **kws):
    ...     return
    >>> FuncSchema.parse(test, strict=False)
    {'b': {'type': <class 'str'>, 'default': <class 'inspect._empty'>}, 'c': {'type': <class 'int'>, 'default': 1}, 'd': {'type': <class 'list'>, 'default': ['d']}, 'e': {'type': <class 'float'>, 'default': 0.1}, 'f': {'type': <class 'set'>, 'default': {'f'}}, 'g': {'type': <class 'tuple'>, 'default': (1, 2)}, 'h': {'type': <class 'bool'>, 'default': True}, 'i': {'type': <class 'set'>, 'default': {1}}}
    >>> def test(a):
    ...     return
    >>> try:FuncSchema.parse(test, strict=True)
    ... except TypeError as e: e
    TypeError('Parameter `a` has no type and no default value.')
    >>> def test(b: str):
    ...     return
    >>> FuncSchema.parse(test, strict=True)
    {'b': {'type': <class 'str'>, 'default': <class 'inspect._empty'>}}
    >>> FuncSchema.parse(test, strict=True, fill_default=True)
    {'b': {'type': <class 'str'>, 'default': ''}}
    >>> def test(**kws):
    ...     return
    >>> try:FuncSchema.parse(test, strict=True)
    ... except TypeError as e: e
    TypeError('Parameter `kws` has no type and no default value.')
    >>> def test(*args):
    ...     return
    >>> try:FuncSchema.parse(test, strict=True)
    ... except TypeError as e: e
    TypeError('Parameter `args` has no type and no default value.')
    >>> FuncSchema.convert("1", int)
    1
    >>> FuncSchema.convert("1", str)
    '1'
    >>> FuncSchema.convert("1", float)
    1.0
    >>> FuncSchema.convert(0, bool)
    False
    >>> FuncSchema.convert('1', bool)
    True
    >>> FuncSchema.convert('[[1, 1]]', dict)
    {1: 1}
    >>> FuncSchema.convert('{"1": "1"}', dict)
    {'1': '1'}
    >>> FuncSchema.convert('[1, 1]', set)
    {1}
    >>> FuncSchema.convert('[1, 1]', tuple)
    (1, 1)
    >>> FuncSchema.convert('[1, "1"]', list)
    [1, '1']
    """

    ALLOW_TYPES = {int, float, str, tuple, list, set, dict, bool}
    JSON_TYPES = {tuple, list, set, dict, bool}

    @classmethod
    def parse(cls, function: Callable, strict=True, fill_default=False):
        sig = inspect.signature(function)
        result = {}
        for param in sig.parameters.values():
            if param.annotation is param.empty:
                if param.default is param.empty:
                    if strict:
                        raise TypeError(
                            f"Parameter `{param.name}` has no type and no default value."
                        )
                    else:
                        continue
                tp = type(param.default)
            else:
                tp = param.annotation
            if tp in cls.ALLOW_TYPES:
                default = param.default
                if fill_default and default is inspect._empty:
                    default = get_type_default(tp, default)
                result[param.name] = {"type": tp, "default": default}
        return result

    @classmethod
    def convert(cls, obj, target_type):
        if isinstance(obj, str) and target_type in cls.JSON_TYPES:
            return target_type(json.loads(obj))
        else:
            return target_type(obj)


class InlinePB(object):
    """Inline progress bar.

    Demo::

        with InlinePB(100) as pb:
            for i in range(100):
                pb.add(1)
                time.sleep(0.03)
        # Progress:  41 / 100  41% [||||||         ] |   33 units/s
        with InlinePB(100) as pb:
            for i in range(1, 101):
                pb.update(i)
                time.sleep(0.03)
        # Progress:  45 / 100  45% [||||||         ] |   33 units/s

    """

    def __init__(
        self,
        total,
        start=0,
        maxlen=50,
        fresh_interval=0.1,
        timer=time.time,
        sig="|",
        sig_len=15,
    ):
        self.total = total
        self.done = start
        self.maxlen = maxlen
        self.cache = deque([], maxlen=maxlen)
        self.timer = timer
        self.last_fresh = self.timer()
        self.fresh_interval = fresh_interval
        self.sig = sig
        self.sig_len = sig_len

    def update(self, done):
        self.add(done - self.done)

    def add(self, num=1):
        self.done += num
        self.cache.append((self.done, self.timer()))
        if self.need_fresh():
            self.fresh()
            self.last_fresh = self.timer()

    def need_fresh(self):
        if self.timer() - self.last_fresh > self.fresh_interval:
            return True
        else:
            return False

    def speed(self) -> int:
        if len(self.cache) > 1:
            a, b = self.cache[0], self.cache[-1]
            return round((b[0] - a[0]) / (b[1] - a[1]))
        elif self.cache:
            return round(self.done / (self.timer() - self.cache[0][1]))
        else:
            return 0

    def __enter__(self):
        self._fill = len(str(self.total))
        self._end = f'{" " * 10}\r'
        return self

    def __exit__(self, *_):
        if not any(_):
            self.fresh()
            print(flush=True)

    def sig_string(self, percent):
        return f"[{self.sig * int(self.sig_len * percent / 100)}{' ' * (self.sig_len - int(self.sig_len * percent / 100))}]"

    def fresh(self):
        percent = int(100 * self.done / self.total)
        done = f"{self.done}".rjust(self._fill, " ")
        print(
            f"Progress: {done} / {self.total} {percent: >3}% {self.sig_string(percent)} | {self.speed(): >4} units/s",
            end=self._end,
            flush=True,
        )


class SizedTimedRotatingFileHandler(TimedRotatingFileHandler):
    """TimedRotatingFileHandler with maxSize, to avoid files that are too large.

    no test.

    Demo::

        import logging
        import time

        logger = logging.getLogger("test")
        h = SizedTimedRotatingFileHandler("test.log", "d", 1, 3, maxBytes=1)
        h.setFormatter(logging.Formatter("%(asctime)s %(levelname)s %(message)s"))
        logger.addHandler(h)

        for i in range(5):
            logger.warning(str(i) * 100)
            time.sleep(1)
        # 2024/06/25 22:47   134     test.log.20240625_224717
        # 2024/06/25 22:47   134     test.log.20240625_224718
        # 2024/06/25 22:47   134     test.log.20240625_224719
    """

    def __init__(
        self,
        filename,
        when="h",
        interval=1,
        backupCount=0,
        maxBytes=0,
        encoding=None,
        delay=False,
        utc=False,
    ):
        """
        Initialize the timed backup file handler.

        :param filename: The name of the log file.
        :param when: The time unit for timed backups, can be "h" (hours) or "d" (days)
        :param interval: The interval for timed backups, with the unit determined by the 'when' parameter
        :param backupCount: The maximum number of backup files to keep
        :param maxBytes: The file size limit before triggering a backup (0 means no limit)
        :param encoding: The encoding of the file
        :param delay: Whether to delay opening the file until the first write
        :param utc: Whether to use UTC time for naming backups
        """
        super().__init__(filename, when, interval, backupCount, encoding, delay, utc)
        self.maxBytes = maxBytes
        self.suffix = "%Y%m%d_%H%M%S"
        self.extMatch = re.compile(
            r"^\d{4}\d{2}\d{2}_\d{2}\d{2}\d{2}(\.\w+)?$", re.ASCII
        )

    def shouldRollover(self, record):
        """
        Determine if rollover should occur.
        Basically, see if the supplied record would cause the file to exceed
        the size limit we have.
        """
        if super().shouldRollover(record):
            return True

        if self.maxBytes > 0:
            if self.stream.tell() >= self.maxBytes:
                return True
        return False

    def doRollover(self):
        """
        Do a rollover, as described by the base class documentation.
        However, also check for the maxBytes parameter and rollover if needed.
        """
        super().doRollover()

        if self.maxBytes > 0:
            self.stream.close()
            self.stream = self._open()


def get_type_default(tp, default=None):
    """Get the default value for a type. {int: 0, float: 0.0, bytes: b"", str: "", list: [], tuple: (), set: set(), dict: {}}"""
    return {
        int: 0,
        float: 0.0,
        bytes: b"",
        str: "",
        list: [],
        tuple: (),
        set: set(),
        dict: {},
    }.get(tp, default)


def func_cmd(function: Callable, run=True, auto_default=False):
    """Handle function with argparse, typing-hint is nessessary.

    Demo::

        def test(str: str, /, int=1, *, list=["d"], float=0.1, set={"f"}, tuple=(1, 2), bool=True, dict={"k": 1}):
            \"\"\"Test demo function.

            Args:
                str (str): str.
                int (int, optional): int. Defaults to 1.
                list (list, optional): list. Defaults to ["d"].
                float (float, optional): float. Defaults to 0.1.
                set (dict, optional): set. Defaults to {"f"}.
                tuple (tuple, optional): tuple. Defaults to (1, 2).
                bool (bool, optional): bool. Defaults to True.
                dict (dict, optional): dict. Defaults to {"k": 1}.
            \"\"\"
            print(locals())

        # raise ValueError if auto_default is False and user do not input nessessary args.
        func_cmd(test, auto_default=False)

        CMD args:

        > python app.py
        ValueError: `str` has no default value.

        > python app.py --str 1 --int 2 --float 1.0 --list "[1,\"a\"]" --tuple "[2,\"b\"]" --set "[1,1,2]" --dict "{\"k\":\"v\"}"
        {'str': '1', 'int': 2, 'list': [1, 'a'], 'float': 1.0, 'set': {1, 2}, 'tuple': (2, 'b'), 'bool': True, 'dict': {'k': 'v'}}

        > python app.py -s 1 -i 2 -f 1.0 -l "[1,\"a\"]" -t "[2,\"b\"]" -s "[1,1,2]" -d "{\"k\":\"v\"}"
        {'str': '[1,1,2]', 'int': 2, 'list': [1, 'a'], 'float': 1.0, 'set': {'f'}, 'tuple': (2, 'b'), 'bool': True, 'dict': {'k': 'v'}}

        > python app.py -h
        usage: Test demo function.

            Args:
                str (str): str.
                int (int, optional): int. Defaults to 1.
                list (list, optional): list. Defaults to ["d"].
                float (float, optional): float. Defaults to 0.1.
                set (dict, optional): set. Defaults to {"f"}.
                tuple (tuple, optional): tuple. Defaults to (1, 2).
                bool (bool, optional): bool. Defaults to True.
                dict (dict, optional): dict. Defaults to {"k": 1}.


        options:
        -h, --help            show this help message and exit
        -s STR, --str STR     {'type': <class 'str'>, 'default': <class 'inspect._empty'>}
        -i INT, --int INT     {'type': <class 'int'>, 'default': 1}
        -l LIST, --list LIST  {'type': <class 'list'>, 'default': ['d']}
        -f FLOAT, --float FLOAT
                                {'type': <class 'float'>, 'default': 0.1}
        -se SET, --set SET    {'type': <class 'set'>, 'default': {'f'}}
        -t TUPLE, --tuple TUPLE
                                {'type': <class 'tuple'>, 'default': (1, 2)}
        -b BOOL, --bool BOOL  {'type': <class 'bool'>, 'default': True}
        -d DICT, --dict DICT  {'type': <class 'dict'>, 'default': {'k': 1}}
    """
    from argparse import ArgumentParser

    parser = ArgumentParser(usage=function.__doc__)
    schema = FuncSchema.parse(function, strict=True, fill_default=False)
    seen_shorten = {"h"}
    for key, value in schema.items():
        args = [f"--{key}"]
        for i in range(1, 4):
            if i > len(key):
                break
            short = key[:i]
            if short not in seen_shorten:
                seen_shorten.add(short)
                args.insert(0, f"-{short}")
                break
        kwargs = dict(dest=key, help=str(value))
        if value["type"] in {int, float, str}:
            kwargs["type"] = value["type"]
        else:
            kwargs["type"] = str
        if value["default"] is inspect._empty:
            kwargs["default"] = get_type_default(value["type"])
        else:
            kwargs["default"] = value["default"]
        parser.add_argument(*args, **kwargs)
    parsed = parser.parse_args()
    kwargs = {}
    args = []
    for key, value in schema.items():
        if value["default"] is inspect._empty:
            if auto_default:
                args.append(FuncSchema.convert(getattr(parsed, key), value["type"]))
            else:
                raise ValueError(f"`{key}` has no default value.")
        else:
            kwargs[key] = FuncSchema.convert(getattr(parsed, key), value["type"])
    if run:
        return function(*args, **kwargs)
    else:
        return args, kwargs


def file_import(file_path, names):
    """Import function from file path.

    Demo::
        >>> from pathlib import Path
        >>> file_path = Path(__file__).parent / "utils.py"
        >>> list(file_import(file_path, ["get_hash", "find_jsons"]).keys())
        ['get_hash', 'find_jsons']
    """
    module_name = os.path.splitext(os.path.basename(file_path))[0]
    spec = importlib.util.spec_from_file_location(module_name, file_path)
    module = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(module)
    return {name: getattr(module, name) for name in names}


class RotatingFileWriter:
    """RotatingFileWriter class for writing to a file with rotation support.

    Demo::

        >>> # test normal usage
        >>> writer = RotatingFileWriter("test.log", max_size=10 * 1024, max_backups=1)
        >>> writer.write("1" * 10)
        >>> writer.path.stat().st_size
        0
        >>> writer.flush()
        >>> writer.path.stat().st_size
        10
        >>> writer.clean_backups(writer.max_backups)
        >>> writer.unlink_file()
        >>> # test rotating
        >>> writer = RotatingFileWriter("test.log", max_size=20, max_backups=2)
        >>> writer.write("1" * 15)
        >>> writer.write("1" * 15)
        >>> writer.write("1" * 15, flush=True)
        >>> writer.path.stat().st_size
        15
        >>> len(writer.backup_path_list())
        2
        >>> writer.clean_backups(writer.max_backups)
        >>> writer.unlink_file()
        >>> # test no backups
        >>> writer = RotatingFileWriter("test.log", max_size=20, max_backups=0)
        >>> writer.write("1" * 15)
        >>> writer.write("1" * 15)
        >>> writer.write("1" * 15, flush=True)
        >>> writer.path.stat().st_size
        15
        >>> len(writer.backup_path_list())
        0
        >>> writer.clean_backups(writer.max_backups)
        >>> writer.unlink_file()
    """

    check_exist_every = 100

    def __init__(
        self,
        path: Union[Path, str],
        max_size=5 * 1024**2,
        max_backups=0,
        encoding="utf-8",
    ):
        if max_backups < 0:
            raise ValueError("max_backups must be greater than -1, 0 for itself.")
        self.path = Path(path)
        self.path.parent.mkdir(parents=True, exist_ok=True)
        self.max_size = max_size
        self.max_backups = max_backups
        self.encoding = encoding
        self.file = self.reopen_file()
        self._check_exist_count = self.check_exist_every + 1
        self._rotate_lock = Lock()

    def unlink_file(self):
        self.close_file()
        self.path.unlink(missing_ok=True)

    def close(self):
        return self.close_file()

    def close_file(self):
        file_obj = getattr(self, "file", None)
        if file_obj and not file_obj.closed:
            file_obj.close()
            self.file = None

    def reopen_file(self):
        self.close_file()
        self.file = self.path.open("a")
        return self.file

    def check_exist(self):
        return not (
            self._check_exist_count > self.check_exist_every and not self.path.is_file()
        )

    def rotate(self, new_length):
        with self._rotate_lock:
            if self.need_rotate(new_length):
                if self.max_backups > 0:
                    self.close_file()
                    now = time.strftime("%Y%m%d%H%M%S")
                    for index in range(self.max_backups):
                        suffix = f"{now}_{index}" if index else now
                        target_path = self.path.with_name(f"{self.path.name}.{suffix}")
                        if target_path.is_file():
                            # already rotated
                            continue
                        else:
                            break
                    else:
                        raise RuntimeError(
                            "max_backups is too small for writing too fast"
                        )
                    self.path.rename(target_path)
                    self.reopen_file()
                    self.clean_backups(count=None)
                else:
                    self.file.seek(0)
                    self.file.truncate()

    def need_rotate(self, new_length):
        return self.max_size and self.file.tell() + new_length > self.max_size

    def ensure_file(self, new_length=0):
        if not self.file:
            self.reopen_file()
        elif not self.check_exist():
            self.reopen_file()
        elif self.need_rotate(new_length):
            self.rotate(new_length)

    def backup_path_list(self):
        return list(self.path.parent.glob(f"{self.path.name}.*"))

    def clean_backups(self, count=None):
        """Clean oldest {count} backups, if count is None, it will clean up to max_backups."""
        path_list = self.backup_path_list()
        if path_list:
            if count is None:
                count = len(path_list) - self.max_backups
            if count > 0:
                path_list.sort(key=lambda x: x.stat().st_mtime)
                for deleted, path in enumerate(path_list, 1):
                    path.unlink(missing_ok=True)
                    if deleted >= count:
                        break

    def flush(self):
        self.file.flush()

    def write(self, text: str, flush=False):
        self._check_exist_count += 1
        self.ensure_file(len(text))
        self.file.write(text)
        if flush:
            self.file.flush()

    def print(self, *strings, end="\n", sep=" ", flush=False):
        text = f"{sep.join(map(str, strings))}{end}"
        self.write(text, flush=flush)

    def __del__(self):
        self.close_file()


def test_bg_task():
    async def _test_bg_task():
        async def coro():
            return True

        task = bg_task(coro())
        assert await task is True
        result = (task.done(), len(_bg_tasks))
        assert result == (True, 0), result

    asyncio.get_event_loop().run_until_complete(_test_bg_task())


def test_named_lock():
    def test_sync():
        import time
        from concurrent.futures import ThreadPoolExecutor
        from threading import Lock, Semaphore

        def _test1():
            with NamedLock("_test1", Lock, timeout=0.05) as lock:
                time.sleep(0.2)
                return bool(lock)

        with ThreadPoolExecutor(10) as pool:
            tasks = [pool.submit(_test1) for _ in range(3)]
            result = [i.result() for i in tasks]
            assert result == [True, False, False], result
        assert len(NamedLock._SYNC_CACHE) == 1
        NamedLock.clear_unlocked()
        assert len(NamedLock._SYNC_CACHE) == 0

        def _test2():
            with NamedLock("_test2", lambda: Semaphore(2), timeout=0.05) as lock:
                time.sleep(0.2)
                return bool(lock)

        with ThreadPoolExecutor(10) as pool:
            tasks = [pool.submit(_test2) for _ in range(3)]
            result = [i.result() for i in tasks]
            assert result == [True, True, False], result

    def test_async():
        import asyncio

        async def main():
            async def _test1():
                async with NamedLock("_test1", asyncio.Lock, timeout=0.05) as lock:
                    await asyncio.sleep(0.2)
                    return bool(lock)

            tasks = [asyncio.create_task(_test1()) for _ in range(3)]
            result = [await i for i in tasks]
            assert result == [True, False, False], result
            assert len(NamedLock._ASYNC_CACHE) == 1
            NamedLock.clear_unlocked()
            assert len(NamedLock._ASYNC_CACHE) == 0

            async def _test2():
                async with NamedLock(
                    "_test2", lambda: asyncio.Semaphore(2), timeout=0.05
                ) as lock:
                    await asyncio.sleep(0.2)
                    return bool(lock)

            tasks = [asyncio.create_task(_test2()) for _ in range(3)]
            result = [await i for i in tasks]
            assert result == [True, True, False], result

        asyncio.get_event_loop().run_until_complete(main())

    test_sync()
    test_async()


def test():
    test_bg_task()
    test_named_lock()


if __name__ == "__main__":
    __name__ = "morebuiltins.functools"
    test()
    import doctest

    doctest.testmod()



import re
from datetime import datetime, timedelta
from typing import Callable, Dict, List, Optional, Tuple, Union, Set, Generator


__all__ = ["ScheduleTimer", "Crontab"]


class ScheduleTimer:
    """The ScheduleTimer class facilitates the creation and evaluation of datetime patterns for scheduling purposes.

    It includes mechanisms to parse patterns involving logical operations (AND, OR) and comparison checks (equality, inequality, arithmetic, and custom range checks).

    Comparison Operators:

        Equality (= or ==): Tests for equality between datetime parts.
            Example: "hour=12" checks if it's exactly 12 o'clock.
        Inequality (!=): Ensures inequality between datetime parts.
            Example: "minute!=30" for minutes not being 30.
        Less Than (<): Requires the left datetime part to be less than the right.
            Example: "day<15" for days before the 15th.
        Less Than or Equal (<=): Allows the left datetime part to be less or equal to the right.
            Example: "month<=6" covers January to June.
        Greater Than (>): Ensures the left datetime part is greater than the right.
            Example: "hour>18" for evenings after 6 PM.
        Greater Than or Equal (>=): Allows the left datetime part to be greater or equal to the right.
            Example: "weekday>=MO" starts from Monday.
        Division Modulo (/): Divisibility check for the left digit by the right digit in datetime format.
            Example: "minute/15" checks for quarter hours.
        Range Inclusion (@): Confirms if the left time falls within any defined ranges in the right string.
            Example: "hour@9-11,13-15" for office hours.

    Logical Operators:
        AND (&): Both conditions must hold true.
            Example: "hour=12&minute=30" for exactly 12:30 PM.
        OR (; or |): At least one of the conditions must be true.
            Example: "hour=12;hour=18" for noon or 6 PM.
        Negation (!): Inverts the truth of the following condition. At the start of the pattern, the condition is negated.
            Example: "!hour=12" excludes noon.

    Demo:

    >>> start_date = datetime.strptime("2023-02-01 00:00:00", "%Y-%m-%d %H:%M:%S")
    >>> list(ScheduleTimer.iter_datetimes("%M=05|%M=15", count=3, start_date=start_date, callback=str))
    ['2023-02-01 00:05:00', '2023-02-01 00:15:00', '2023-02-01 01:05:00']
    >>> list(ScheduleTimer.iter_datetimes("%H:%M=15:30", count=3, start_date=start_date, callback=str))
    ['2023-02-01 15:30:00', '2023-02-02 15:30:00', '2023-02-03 15:30:00']
    >>> list(ScheduleTimer.iter_datetimes("%H:%M=15:30&%d=15", count=3, start_date=start_date, callback=str))
    ['2023-02-15 15:30:00', '2023-03-15 15:30:00', '2023-04-15 15:30:00']
    >>> list(ScheduleTimer.iter_datetimes("%H:%M=15:30&%d>=15", count=3, start_date=start_date, callback=str))
    ['2023-02-15 15:30:00', '2023-02-16 15:30:00', '2023-02-17 15:30:00']
    >>> list(ScheduleTimer.iter_datetimes("%M@15-16&%d>=15", count=3, start_date=start_date, callback=str))
    ['2023-02-15 00:15:00', '2023-02-15 00:16:00', '2023-02-15 01:15:00']
    >>> list(ScheduleTimer.iter_datetimes("%M/15", count=5, start_date=start_date, callback=str))
    ['2023-02-01 00:00:00', '2023-02-01 00:15:00', '2023-02-01 00:30:00', '2023-02-01 00:45:00', '2023-02-01 01:00:00']
    """

    prefix_not = "!"
    split_or = re.compile(";|\|")
    split_and = re.compile("&")
    split_calc = re.compile("(=+|!=|<=?|>=?|@|/)")
    magic_methods = {
        "=": "__eq__",
        "==": "__eq__",
        "!=": "__ne__",
        "<": "__lt__",
        "<=": "__le__",
        ">": "__gt__",
        ">=": "__ge__",
    }
    custom_methods = {"/": "divide", "@": "between"}
    step_choices = [
        ("%S", timedelta(seconds=1)),
        ("%M", timedelta(minutes=1)),
        ("%H", timedelta(hours=1)),
    ]
    min_delta = timedelta(minutes=1)

    def __init__(self):
        raise NotImplementedError(
            "Performance issue needs to be optimized, use the classmethod instead"
        )

    @staticmethod
    def divide(a: str, b: str):
        if a.isdigit() and b.isdigit():
            return int(a) % int(b) == 0
        else:
            return False

    @staticmethod
    def between(a: str, b: str):
        # many ranges
        range_regex = re.compile("^(\d+)-(\d+)$")
        for time_range in b.split(","):
            if a == time_range:
                return True
            else:
                match = range_regex.match(time_range)
                if match and int(a) >= int(match[1]) and int(a) <= int(match[2]):
                    return True
        return False

    @classmethod
    def get_step(cls, pattern: str):
        for step, delta in cls.step_choices:
            if step in pattern:
                if delta < cls.min_delta:
                    raise ValueError(
                        f"Step delta must be greater than {cls.min_delta.total_seconds()}, but given is {step}, or increase cls.min_delta but the performance will be worse"
                    )
                return timedelta(minutes=1)
        return timedelta(days=1)

    @classmethod
    def sort_pattern(cls, pattern: str):
        steps: Dict[str, timedelta] = {}
        for or_part in cls.split_or.split(pattern):
            and_parts = []
            for and_part in cls.split_and.split(or_part):
                steps[and_part] = cls.get_step(and_part)
                and_parts.append(and_part)
            and_parts.sort(key=lambda x: steps[x], reverse=True)
        return pattern, steps

    @classmethod
    def iter_datetimes(
        cls,
        pattern: str,
        count: int = 1,
        start_date: Optional[datetime] = None,
        max_tries: int = 1000000,
        callback: Callable = lambda x: x,
        yield_tries: bool = False,
    ) -> Generator:
        if start_date is None:
            _start_date = datetime.now().replace(microsecond=0)
        else:
            _start_date = start_date
        pattern_list, steps = cls.sort_pattern(pattern)
        min_step = min(steps.values())
        last_item = None
        matched = 0
        for tries in range(1, max_tries + 1):
            ok, miss_pattern = cls.is_at(pattern_list, _start_date)
            if ok:
                item = callback(_start_date)
                if not last_item or item != last_item:
                    matched += 1
                    if yield_tries:
                        yield tries, item
                    else:
                        yield item
                    if matched == count:
                        break
                step = min_step
            else:
                step = steps[miss_pattern]
            _start_date += step

    @classmethod
    def is_at(cls, pattern: Union[list, str], date: datetime) -> Tuple[bool, str]:
        if isinstance(pattern, str):
            pattern_list, _ = cls.sort_pattern(pattern)
        else:
            pattern_list = pattern
        miss_pattern: str = ""
        ok = False
        for or_pattern in cls.split_or.split(pattern_list):
            for and_pattern in cls.split_and.split(or_pattern):
                if and_pattern[0] == cls.prefix_not:
                    and_pattern = and_pattern[1:]
                    be_not = True
                else:
                    be_not = False
                _a, cmp, _b = cls.split_calc.split(and_pattern)
                if _a and _b:
                    a = date.strftime(_a)
                    b = date.strftime(_b)
                    magic_method = getattr(a, cls.magic_methods.get(cmp, ""), None)
                    if magic_method:
                        result = magic_method(b)
                    else:
                        result = getattr(cls, cls.custom_methods[cmp])(a, b)
                    if be_not:
                        if not result:
                            continue
                        else:
                            # no match break
                            miss_pattern = miss_pattern or and_pattern
                            break
                    elif result:
                        continue
                    else:
                        # no match break
                        miss_pattern = miss_pattern or and_pattern
                        break
                else:
                    # no match break
                    miss_pattern = miss_pattern or and_pattern
                    break
            else:
                # each and_pattern is hit
                ok = True
                break
        return ok, miss_pattern


class Crontab:
    """Crontab python parser.

    Demo:

    >>> start_date = datetime.strptime("2023-02-01 00:00:00", "%Y-%m-%d %H:%M:%S")
    >>> list(Crontab.iter_datetimes("*/15 * * * *", count=10, start_date=start_date, callback=lambda i: i.strftime("%M")))
    ['00', '15', '30', '45', '00', '15', '30', '45', '00', '15']
    >>> list(Crontab.iter_datetimes("* * * * 2,4,6", count=10, start_date=start_date, callback=lambda i: i.strftime("%a")))
    ['Thu', 'Sat', 'Tue', 'Thu', 'Sat', 'Tue', 'Thu', 'Sat', 'Tue', 'Thu']
    >>> list(Crontab.iter_datetimes("0 0 11-19/4,8,30 * *", count=10, start_date=start_date, callback=lambda i: i.strftime("%m-%d")))
    ['02-08', '02-11', '02-15', '02-19', '03-08', '03-11', '03-15', '03-19', '03-30', '04-08']
    >>> list(Crontab.iter_datetimes("* * * * *", count=1, start_date=start_date))
    [datetime.datetime(2023, 2, 1, 0, 0)]
    >>> list(Crontab.iter_datetimes("5 4-5,6-9/2 5,6 * 3,5", count=3, start_date=start_date, callback=str))
    ['2023-04-05 04:05:00', '2023-04-05 05:05:00', '2023-04-05 06:05:00']
    >>> list(Crontab.iter_datetimes("0 0 1 8,9,10 *", count=3, start_date=start_date, callback=str, yield_tries=True))
    [(7, '2023-08-01 00:00:00'), (120, '2023-09-01 00:00:00'), (232, '2023-10-01 00:00:00')]
    >>> list(Crontab.iter_datetimes("0 0 1 11 *", count=3, start_date=start_date, callback=str, yield_tries=True))
    [(10, '2023-11-01 00:00:00'), (133, '2024-11-01 00:00:00'), (256, '2025-11-01 00:00:00')]
    """

    mins = list(range(0, 60))
    hours = list(range(0, 24))
    days = list(range(1, 32))
    months = list(range(1, 13))
    # crontab 0~7
    weeks = list(range(0, 8))
    # crontab: 0, 7 is Sunday
    # python: Monday is 0 and Sunday is 6
    wrap_cron_to_python = {0: 6, 1: 0, 2: 1, 3: 2, 4: 3, 5: 4, 6: 5, 7: 6}

    @classmethod
    def parse_field(cls, value: str, range_list: List[int]) -> Set[int]:
        """Parse a single cron field and return a list of matching values."""
        if value == "*":
            return set(range_list)
        results: set = set()
        valid_values = set(range_list)
        for val in value.split(","):
            if val == "*":
                results.update(range_list)
                break
            elif val.isdigit():
                results.add(int(val))
                continue
            if "/" in val:
                val, right = val.split("/")
                step = int(right)
            else:
                step = 1
            if val == "*":
                results.update(range(range_list[0], range_list[-1] + 1, step))
            elif "-" in val:
                start, end = map(int, val.split("-"))
                results.update(range(start, end + 1, step))
        return {i for i in results if i in valid_values}

    @classmethod
    def iter_datetimes(
        cls,
        pattern: str,
        count: int = 1,
        start_date: Optional[datetime] = None,
        max_tries: int = 1000000,
        callback: Callable = lambda x: x,
        yield_tries=False,
    ) -> Generator:
        if start_date is None:
            _start_date = datetime.now().replace(microsecond=0)
        else:
            _start_date = start_date
        _start_date = _start_date.replace(second=0)
        M, H, d, m, w = pattern.split(" ")
        mins = cls.parse_field(M, cls.mins)
        hours = cls.parse_field(H, cls.hours)
        days = cls.parse_field(d, cls.days)
        months = cls.parse_field(m, cls.months)
        weeks = cls.parse_field(w, cls.weeks)
        py_weekdays = {cls.wrap_cron_to_python[i] for i in weeks}
        tries = 0
        matched = 0
        last_match = None
        while tries < max_tries:
            tries += 1
            if _start_date.month not in months:
                # fast forward to next month with performance improvement
                temp = _start_date.replace(day=28, hour=0, minute=0) + timedelta(days=4)
                _start_date = temp.replace(day=1)
            elif (
                _start_date.weekday() not in py_weekdays or _start_date.day not in days
            ):
                _start_date = _start_date.replace(hour=0, minute=0) + timedelta(days=1)
            elif _start_date.hour not in hours:
                _start_date = _start_date.replace(minute=0) + timedelta(hours=1)
            elif _start_date.minute not in mins:
                _start_date += timedelta(minutes=1)
            else:
                item = callback(_start_date)
                if not last_match or item != last_match:
                    matched += 1
                    if yield_tries:
                        yield tries, item
                    else:
                        yield item
                    last_match = item
                    if matched == count:
                        break
                _start_date += timedelta(minutes=1)


if __name__ == "__main__":
    import doctest

    doctest.testmod()



__version__ = "1.1.4"
__all__ = [
    "morebuiltins.utils",
    "morebuiltins.date",
    "morebuiltins.functools",
    "morebuiltins.ipc",
    "morebuiltins.request",
    "morebuiltins.download_python",
    "morebuiltins.tk",
    "morebuiltins.emails",
    "morebuiltins.cmd.log_server",
    "morebuiltins.cmd.proxy_checker",
]



import doctest
import importlib
import sys
import time

sys.path.insert(0, "./")


def zipapp_module():
    from morebuiltins.zipapps.main import create_app

    create_app(
        "./morebuiltins",
        output="./morebuiltins.pyz",
        compressed=True,
    )


def test_all():
    # test local usage and pyz usage
    for path in ["./", "./morebuiltins.pyz"]:
        sys.modules.pop("morebuiltins")
        sys.path.insert(0, path)

        import morebuiltins

        assert path.replace("./", "") in morebuiltins.__file__, morebuiltins.__file__

        for name in morebuiltins.__all__:
            module = importlib.import_module(name)
            print(
                time.strftime("%Y-%m-%d %H:%M:%S"),
                "[TEST]",
                module.__name__,
                flush=True,
            )
            if hasattr(module, "test"):
                module.test()
            else:
                result = doctest.testmod(module)
                if result.failed:
                    raise RuntimeError
            print(
                time.strftime("%Y-%m-%d %H:%M:%S"),
                "[PASS]",
                module.__name__,
                flush=True,
            )
            time.sleep(1)
        print("all test ok", flush=True)


def make_docs():
    import importlib
    import re
    from pathlib import Path

    import morebuiltins

    short_doc = ""
    with open("doc.md", "w", encoding="u8") as f:
        for index1, name in enumerate(morebuiltins.__all__, 1):
            module = importlib.import_module(name)
            name1 = f"## {index1}. {name}\n\n"
            short_doc += name1
            print(name1, file=f)
            for index2, name in enumerate(module.__all__, 1):
                member = vars(module)[name]
                doc = member.__doc__
                if doc:
                    doc = doc.lstrip()
                    title = f"{index1}.{index2} `{name}`"
                    lines = re.split("\n", doc, maxsplit=1)
                    head = "%s - %s" % (title, lines[0])
                    short_doc += head + "\n\n"
                    if len(lines) > 1:
                        tail = lines[1]
                        tail = f'```python\n{tail}\n```'
                    else:
                        tail = ""
                    line = f"\n{head}\n\n\n{tail}\n\n"
                    print(
                        line,
                        "\n---\n\n",
                        sep="",
                        file=f,
                    )
            short_doc += "\n"
    path = Path("README.md")
    old_readme = path.read_text("u8")
    start = old_readme.find("<!-- start -->\n") + len("<!-- start -->\n")
    end = old_readme.find("<!-- end -->")
    path.write_text(old_readme[:start] + short_doc + old_readme[end:], encoding="u8")


def main():
    make_docs()
    zipapp_module()
    test_all()


if __name__ == "__main__":
    main()



import builtins
import tkinter as tk
import tkinter.messagebox
from threading import Thread
from tkinter import scrolledtext
from typing import Any, Dict

__all__ = ["TKit"]


class TKit(tk.Tk):
    r"""Tkinter kit for dialog usages.
    Demo::

        def examples():
            while True:
                TKit.ask(0, "0")
                TKit.ask(1, "1")
                TKit.ask(2, "2")
                if TKit.ask(True, "Choose NO", default="no") is True:
                    TKit.ask(0, "Wrong choice")
                    continue
                if (
                    TKit.ask((["1"], ["2", "3"], "4", ["5", "6", "7"]), message="Choose 3:")
                    != "3"
                ):
                    TKit.ask(1, "Wrong choice")
                    continue
                if TKit.ask(
                    [["1"], ["2", "3"], "4", ["5", "6", "7"]],
                    message="Choose 3 and 6:",
                    width=400,
                ) != ["3", "6"]:
                    TKit.ask(2, "Wrong choice")
                    continue
                result = TKit.ask("Input text 1 (press Enter to submit):")

                if result != "1":
                    TKit.ask(2, "Wrong text %s" % repr(result))
                    continue
                result = TKit.ask("Input text 1\\n (press Ctrl-Enter to submit):", textarea=1)

                if result != "1\n":
                    TKit.ask(2, "Wrong text %s" % repr(result))
                    continue

                def test_text(flush=False):
                    import time

                    for i in range(50):
                        print(f"Test print flush={flush} -- {i}", flush=flush)
                        time.sleep(0.02)
                    return "OK"

                with TKit.text_context(
                    test_text,
                    flush=True,
                    __resize_kwargs={"title": "The Title", "toolwindow": True},
                    __text_kwargs={"font": "_ 15"},
                ) as result:
                    TKit.info("result=%s" % result)

                with TKit.text_context(
                    test_text,
                    flush=False,
                    __resize_kwargs={"title": "The Title", "toolwindow": True},
                    __text_kwargs={"font": "_ 15"},
                ) as result:
                    TKit.warn("result=%s" % result)
                break

        examples()
    """

    DEFAULT_WIDTH_RATIO = 4
    DEFAULT_HEIGHT_RATIO = 4

    def resize(
        self,
        width=0,
        height=0,
        pos=(None, None),
        topmost=False,
        alpha=1.0,
        fullscreen=False,
        toolwindow=False,
        disabled=False,
    ):
        return self.resize_tk(
            self,
            width=width,
            height=height,
            pos=pos,
            topmost=topmost,
            alpha=alpha,
            fullscreen=fullscreen,
            toolwindow=toolwindow,
            disabled=disabled,
        )

    @classmethod
    def resize_tk(
        cls,
        tk_obj,
        width=0,
        height=0,
        pos=(None, None),
        topmost=False,
        alpha=1.0,
        fullscreen=False,
        toolwindow=False,
        disabled=False,
        title=None,
    ):
        """Resize and Move the window to the given position, default to the center of the screen."""
        screen_width = tk_obj.winfo_screenwidth()
        screen_height = tk_obj.winfo_screenheight()
        width = width or (screen_width // cls.DEFAULT_WIDTH_RATIO)
        height = height or (screen_height // cls.DEFAULT_HEIGHT_RATIO)
        x, y = pos
        if x is None or y is None:
            x = (screen_width // 2) - (width // 2)
            y = (screen_height // 2) - (height // 2)
        tk_obj.geometry(f"{width}x{height}+{x}+{y}")
        tk_obj.attributes("-topmost", topmost)
        tk_obj.attributes("-alpha", alpha)
        tk_obj.attributes("-fullscreen", fullscreen)
        tk_obj.attributes("-toolwindow", toolwindow)
        tk_obj.attributes("-disabled", disabled)
        if title is not None:
            tk_obj.title(title)

    @classmethod
    def ask(cls, arg: Any, message="", title="", **kwargs):
        if isinstance(arg, bool) or arg is bool:
            return cls.ask_yesno(arg, message, title, **kwargs)
        elif isinstance(arg, int):
            return cls.ask_msgbox(arg, message, title, **kwargs)
        elif isinstance(arg, tuple):
            return cls.ask_choice(arg, message, title, **kwargs)
        elif isinstance(arg, list):
            return cls.ask_checkbox(arg, message, title, **kwargs)
        elif isinstance(arg, str):
            return cls.ask_text(arg, message, title, **kwargs)
        else:
            raise TypeError()

    @classmethod
    def info(cls, message="", title="", **kwargs):
        return cls.ask_msgbox(0, message, title, **kwargs)

    @classmethod
    def warn(cls, message="", title="", **kwargs):
        return cls.ask_msgbox(1, message, title, **kwargs)

    @classmethod
    def error(cls, message="", title="", **kwargs):
        return cls.ask_msgbox(2, message, title, **kwargs)

    @classmethod
    def ask_msgbox(cls, arg: int, message="", title="", **kwargs):
        methods = (
            tkinter.messagebox.showinfo,
            tkinter.messagebox.showwarning,
            tkinter.messagebox.showerror,
        )
        try:
            root = tk.Tk()
            root.withdraw()
            root.attributes("-topmost", True)
            root.focus_force()
            return methods[arg](title=title, message=message)
        except IndexError:
            return methods[-1](title=title, message=message)
        finally:
            root.destroy()

    @classmethod
    def ask_yesno(cls, arg: bool, message="Confirm(YES/NO)", title="", **kwargs):
        # ask YES/NO
        try:
            root = tk.Tk()
            root.withdraw()
            root.attributes("-topmost", True)
            root.focus_force()
            return tkinter.messagebox.askyesno(
                title=title, message=message, default=kwargs.get("default")
            )
        finally:
            root.destroy()

    @classmethod
    def ask_choice(cls, arg: tuple, message="Make your choice", title="", **kwargs):
        # Choose one
        root = cls()
        root.title(title)
        root.resize(topmost=True, toolwindow=True, **kwargs)
        results = [None]
        max_col = max([len(i) for i in arg if isinstance(i, (tuple, list))])
        tk.Label(root, text=message, background="#ffffff").grid(
            row=0, column=0, sticky="ewns", columnspan=max_col
        )
        root.grid_rowconfigure(0, weight=1)
        root.grid_columnconfigure(0, weight=1)

        def cb(result):
            def add_result():
                results.append(result)
                root.destroy()

            return add_result

        for row, texts in enumerate(arg, 1):
            root.grid_rowconfigure(row, weight=1)
            if not isinstance(texts, (tuple, list)):
                texts = [texts]
            for col, text in enumerate(texts, 0):
                root.grid_columnconfigure(col, weight=1)
                b = tk.Button(
                    root,
                    text=str(text),
                    borderwidth=1,
                    relief="ridge",
                    command=cb(text),
                )
                b.grid(row=row, column=col, sticky="ewns")
                if (row, col) == (1, 0):
                    b.focus_force()
        root.mainloop()
        return results[-1]

    @classmethod
    def ask_checkbox(cls, arg: list, message="Make your choices", title="", **kwargs):
        # Choose many
        root = cls()
        root.title(title)
        root.resize(topmost=True, toolwindow=True, **kwargs)
        max_col = max([len(i) for i in arg if isinstance(i, (tuple, list))])
        tk.Label(root, text=message, background="#ffffff").grid(
            row=0, column=0, sticky="ewns", columnspan=max_col
        )
        root.grid_columnconfigure(0, weight=1)
        results: Dict[str, tk.IntVar] = {}
        for row, texts in enumerate(arg, 1):
            root.grid_rowconfigure(row, weight=1)
            if not isinstance(texts, (tuple, list)):
                texts = [texts]
            for col, text in enumerate(texts, 0):
                root.grid_columnconfigure(col, weight=1)
                c = tk.Checkbutton(
                    root,
                    text=str(text),
                    variable=results.setdefault(text, tk.IntVar()),
                    borderwidth=1,
                    relief="ridge",
                )
                c.grid(row=row, column=col, sticky="ewns")
                if (row, col) == (1, 0):
                    c.focus_force()
        tk.Button(
            root,
            text="Submit",
            background="#ffffff",
            command=lambda: root.destroy(),
            height=2,
        ).grid(row=len(arg) + 1, column=0, columnspan=max_col, sticky="ewns")
        root.mainloop()
        return sorted((k for k, v in results.items() if v.get()))

    @classmethod
    def ask_text(cls, arg: str, message="", title="", **kwargs):
        textarea = bool(kwargs.pop("textarea", None))
        message = message or arg
        root = cls()
        root.title(title)
        root.resize(topmost=True, toolwindow=True, **kwargs)

        tk.Label(root, text=message, background="#ffffff").pack(
            expand=True, fill="both"
        )

        text_var = tk.StringVar()

        if textarea:
            text_box: Any = tk.Text(root, height=2)

            def submit(event=None):
                # remove \n
                text = text_box.get("1.0", tk.END)
                if text[-1] == "\n":
                    text = text[:-1]
                text_var.set(text)
                root.destroy()

        else:
            text_box = tk.Entry(root, textvariable=text_var)

            def submit(event=None):
                root.destroy()

            text_box.bind("<Return>", submit)
        text_box.bind("<Control-Return>", submit)
        text_box.pack(expand=True, fill="both")
        text_box.focus_force()
        tk.Button(root, text="Submit", background="#ffffff", command=submit).pack(
            expand=True, fill="both"
        )
        root.mainloop()
        return text_var.get()

    @classmethod
    def text_context(cls, function, *args, **kwargs):
        return TextWindow(function, *args, **kwargs)


class TextWindow:
    def __init__(self, function, *args, **kwargs):
        self.print = print
        self.function = function
        self.args = args
        self.kwargs = kwargs

        self.result = None
        self.error = None
        self._shutdown = False

    def print_text(self, *args, end="\n", sep=" ", file=None, flush=False):
        if flush:
            self.text_box.delete("1.0", tk.END)
        self.text_box.insert(tk.END, f"{sep.join(map(str, args))}{end}")
        self.text_box.see(tk.END)

    def patch_print(self):
        builtins.print = self.print_text

    def restore_print(self):
        if builtins.print is not self.print:
            builtins.print = self.print

    def run(self):
        try:
            self.result = self.function(*self.args, **self.kwargs)
        except Exception as e:
            self.error = e
        finally:
            self.shutdown()

    def __enter__(self):
        try:
            self.root = tk.Tk()
            __resize_kwargs = self.kwargs.pop("__resize_kwargs", {})
            TKit.resize_tk(self.root, **__resize_kwargs)
            __text_kwargs = self.kwargs.pop("__text_kwargs", {})
            self.text_box = scrolledtext.ScrolledText(self.root, **__text_kwargs)
            self.text_box.pack(expand=True, fill="both")
            self.text_box.focus_force()
            self.patch_print()
            t = Thread(target=self.run, daemon=True)
            t.start()
            self.root.mainloop()
        finally:
            self.shutdown()
            if self.error:
                raise self.error

        return self.result

    def shutdown(self):
        try:
            if not self._shutdown:
                self._shutdown = True
                self.restore_print()
                self.root.quit()
        except (tk.TclError, AttributeError):
            pass

    def __exit__(self, *_):
        self.root.withdraw()
        self.shutdown()


def examples():
    while True:
        TKit.ask(0, "0")
        TKit.ask(1, "1")
        TKit.ask(2, "2")
        if TKit.ask(True, "Choose NO", default="no") is True:
            TKit.ask(0, "Wrong choice")
            continue
        if (
            TKit.ask((["1"], ["2", "3"], "4", ["5", "6", "7"]), message="Choose 3:")
            != "3"
        ):
            TKit.ask(1, "Wrong choice")
            continue
        if TKit.ask(
            [["1"], ["2", "3"], "4", ["5", "6", "7"]],
            message="Choose 3 and 6:",
            width=400,
        ) != ["3", "6"]:
            TKit.ask(2, "Wrong choice")
            continue
        result = TKit.ask("Input text 1 (press Enter to submit):")
        if result != "1":
            TKit.ask(2, "Wrong text %s" % repr(result))
            continue
        result = TKit.ask("Input text 1\\n (press Ctrl-Enter to submit):", textarea=1)
        if result != "1\n":
            TKit.ask(2, "Wrong text %s" % repr(result))
            continue

        def test_text(flush=False):
            import time

            for i in range(50):
                print(f"Test print flush={flush} -- {i}", flush=flush)
                time.sleep(0.02)
            return "OK"

        with TKit.text_context(
            test_text,
            flush=True,
            __resize_kwargs={"title": "The Title", "toolwindow": True},
            __text_kwargs={"font": "_ 15"},
        ) as result:
            TKit.info("result=%s" % result)

        with TKit.text_context(
            test_text,
            flush=False,
            __resize_kwargs={"title": "The Title", "toolwindow": True},
            __text_kwargs={"font": "_ 15"},
        ) as result:
            TKit.warn("result=%s" % result)
        break


if __name__ == "__main__":
    examples()
